from numba.core.types import *
from numba.experimental import jitclass

from wrpll_simulation.config import PI_Config


@jitclass
class Phase_Accumlator(object):
    last_o: int8
    o: int8
    freq: float64
    phase: float64

    def __init__(self, freq: float64, phase: float64):
        self.last_o = 0
        self.o = 0
        self.freq = freq
        self.phase = phase

    def update(self, time_increment: float64):
        self.last_o = self.o

        self.phase = (self.phase + 360 * self.freq * time_increment) % 360

        # square wave function
        if self.phase < 180:
            self.o = 0
        else:
            self.o = 1

    def set_freq(self, freq: float64):
        self.freq = freq

    def set_phase(self, phase: float64):
        self.phase = phase

    def is_rising(self) -> bool:
        return not self.last_o and self.o


# GATEWARE
@jitclass
class DDMTD(object):
    FF: int8
    beating: int8
    last_beating: int8

    tag: int32
    tag_ready: int8
    blind_counter: int16
    blind_period: int16

    def __init__(self, blind_period: int16):
        # back to back Flip Flop
        self.FF = 0
        self.beating = 0
        self.last_beating = 0

        # deglticher
        self.tag = 0
        self.tag_ready = 0
        self.blind_counter = 0
        self.blind_period = blind_period

    def sync_update(self, D_in: int8, counter: int32):
        self.last_beating = self.beating

        # FF shifting
        self.beating = self.FF
        self.FF = D_in

        self.deglitcher_first_edge(counter)

    def deglitcher_first_edge(self, counter: int32):
        if self.blind_counter == 0 and self.beating_is_rising():
            self.tag = counter
            self.tag_ready = 1
        else:
            self.tag_ready = 0

        if self.beating:
            self.blind_counter = self.blind_period - 1

        if self.blind_counter != 0:
            self.blind_counter -= 1

    def beating_is_rising(self) -> bool:
        return not self.last_beating and self.beating


@jitclass
class EventManager_IRQ(object):
    trigger: int8
    tag_csr: int32

    trigger_index: int64

    def __init__(self):
        self.trigger = 0
        self.tag_csr = 0

        # for simulating delay
        self.trigger_index = 0

    def sync_update(self, index):
        if self.trigger:
            self.trigger = 0
            self.trigger_index = index

    def multireg(self, trigger: int8, tag: int32):
        if trigger:
            self.trigger = 1
            self.tag_csr = tag

    def is_due(self, index, delay):
        return index - self.trigger_index == delay


# FIRMWARE
@jitclass
class Tag_Collector(object):
    setpt_beating_period: int64

    gtx_tag_ready: bool
    gtx_tag: int32
    main_tag_ready: bool
    main_tag: int32

    def __init__(self, setpt_beating_period: int64):
        self.setpt_beating_period = setpt_beating_period

        # for main PLL
        self.gtx_tag_ready = False
        self.gtx_tag = 0
        self.main_tag_ready = False
        self.main_tag = 0

    def collect_gtx_tag(self, tag: int32):
        self.gtx_tag = tag
        self.gtx_tag_ready = True

    def collect_main_tag(self, tag: int32):
        self.main_tag = tag
        self.main_tag_ready = True

    def get_period_error(self) -> int32:
        # period_error = n * ideal_beating - GTX_tag(n)
        period_error = (-self.gtx_tag) % self.setpt_beating_period

        if period_error > self.setpt_beating_period / 2:
            return period_error - self.setpt_beating_period

        return period_error

    def get_phase_error(self) -> int32:
        # tag_diff = main_tag(n) - gtx_tag(n)
        tag_diff = (self.main_tag - self.gtx_tag) % self.setpt_beating_period

        # mapping tags from [0, 2π] -> [-π, π]
        if tag_diff > self.setpt_beating_period / 2:
            return tag_diff - self.setpt_beating_period

        return tag_diff

    def set_phase_error_ready(self, ready: bool):
        self.main_tag_ready = ready
        self.gtx_tag_ready = ready

    def is_phase_error_ready(self) -> bool:
        return self.main_tag_ready and self.gtx_tag_ready


@jitclass
class PI_loop(object):
    KP: int64
    KI: int64
    integrator: int64

    center_freq: float64
    adpll: int32
    base_adpll: int32
    adpll_limit: int32

    i2c_transfer_index: int64

    def __init__(
        self,
        PI_Conifg: PI_Config,
        center_freq: float64,
        base_adpll: int32,
        adpll_limit: int32,
    ):
        # PI controller
        self.KP = PI_Conifg.KP
        self.KI = PI_Conifg.KI
        self.integrator = 0

        # ADPLL calcuation
        self.center_freq = center_freq
        self.adpll = base_adpll
        self.base_adpll = base_adpll
        self.adpll_limit = adpll_limit

        # for simulating delay
        self.i2c_transfer_index = 0

    def update(self, index, tag_error: int32):
        self.i2c_transfer_index = index
        self.adpll = self.get_adpll(tag_error)

    def get_adpll(self, tag_error: int32) -> int32:
        prop = tag_error * self.KP
        self.integrator += tag_error * self.KI
        return self.cramp_adpll(self.base_adpll + int(prop + self.integrator))

    def get_new_freq(self):
        return self.center_freq * (1 + self.adpll * 0.0001164 / 1_000_000)

    def cramp_adpll(self, adpll: int32) -> int32:
        return max(min(self.adpll_limit, adpll), -self.adpll_limit)

    def i2c_is_due(self, index, delay):
        return index - self.i2c_transfer_index == delay
